//-----------------------------------------------------------------------------
//
//
// Direct Sampler : gyabo
//
//
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <vector>
#include <windows.h>
#include <process.h>
#include <mmsystem.h>

#include <dsound.h>
#pragma comment(lib, "winmm.lib")
#pragma comment(lib, "user32.lib")
#pragma comment(lib, "ole32.lib")
#pragma comment(lib, "gdi32.lib")
#pragma comment(lib, "dxguid.lib")
#pragma comment(lib, "dsound.lib")

#define SND_CAPS        DSBCAPS_LOCSOFTWARE   | DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLPAN | \
                        DSBCAPS_CTRLVOLUME    | DSBCAPS_GLOBALFOCUS   | DSBCAPS_CTRLPOSITIONNOTIFY | \
                        DSBCAPS_GETCURRENTPOSITION2
#define RELEASE(x)      { if(x) (x)->Release(); x = NULL; }


//-----------------------------------------------------------------------------
// Direct Sampler
//-----------------------------------------------------------------------------
struct DirectSampler {
  enum {
    MAX_CH      = 8,
    MAX_SAMPLES = 256,
    PITCH_BASE  = 400, //tekitou
  };
  LPDIRECTSOUND8       dsnd;
  LPDIRECTSOUNDBUFFER  dsb[MAX_SAMPLES][MAX_CH];
  DWORD                NumSamples;
  
  void Term() {
    for(int i = 0 ; i < MAX_SAMPLES; i++)
      for(int ch = 0 ; ch < MAX_CH; ch++)
        RELEASE(dsb[i][ch]);
    RELEASE(dsnd);
  }
  
  DirectSampler() {
    NumSamples = 0;
    dsnd = NULL;
    memset(dsb, 0, sizeof(dsb));
  }
  
  ~DirectSampler() { Term(); }
  
  //CREATE FILE -> cat wav1.wav wav2.wav .... wavN.wav > src
  HRESULT Init(HWND hWnd, void *src, int size) {
    HRESULT hRet = E_FAIL;
    do {
      if(hRet = DirectSoundCreate8(NULL, &dsnd, NULL)) {
        printf("Failed DirectSoundCreate8:%08X\n", hRet);
        break;
      }
      if(hRet = dsnd->SetCooperativeLevel(hWnd, DSSCL_NORMAL)) {
        printf("Failed DirectSound SetCooperativeLevel:%08X\n", hRet);
        break;
      }
      
      //Read cat samples
      int Index = 0;
      unsigned char *source = (unsigned char *)src;
      unsigned char *endptr = source + size;
      unsigned char *cur    = source;
      BOOL done = FALSE;
      while(Index < MAX_SAMPLES) {
        while(*(unsigned long *)cur != FOURCC_RIFF) {
          if(cur >= endptr) {
            done = TRUE;
            break;
          }
          cur++;
        } //while
        if(done) break;
        
        //Create Buffer(http://www.kk.iij4u.or.jp/~kondo/wave/, https://ccrma.stanford.edu/courses/422/projects/WaveFormat/)
        LPDIRECTSOUNDBUFFER  lpRet = NULL;
        WAVEFORMATEX        *lpwfx = (WAVEFORMATEX *)&cur[20];
        LPVOID               data  = (LPVOID)&cur[44];
        DWORD                size  = *(DWORD *)&cur[40];
        DSBUFFERDESC         dsbd  = { sizeof(DSBUFFERDESC), SND_CAPS, size, 0, lpwfx, GUID_NULL };
        if(hRet = dsnd->CreateSoundBuffer(&dsbd, &lpRet, NULL)) {
          printf("Failed DirectSound CreateSoundBuffer:%08X", hRet);
          break;
        }
        
        //Fill Buffer
        DWORD  dwSize1, dwSize2;
        LPVOID lpDest1 = NULL, lpDest2 = NULL;
        if(lpRet->Lock(0, size, &lpDest1, &dwSize1, &lpDest2, &dwSize2, 0) == DS_OK) {
          if(lpDest1 && dwSize1) memcpy(lpDest1, data, dwSize1);
          //if(lpDest2 && dwSize2) memcpy(lpDest2, data, dwSize2); //shiran
          lpRet->Unlock(lpDest1, dwSize1, lpDest2, dwSize2);
        }
        
        //duplicate
        dsb[Index][0] = lpRet;
        for(int i = 1; i < MAX_CH; i++)
          dsnd->DuplicateSoundBuffer(dsb[Index][0], &dsb[Index][i]);
        
        //next wave(tekitou)
        cur += size;
        Index++;
        NumSamples++;
      } //for
    } while(0);
    return hRet;
  }
  
  void Play(DWORD ch, DWORD inst, int pitch, BOOL IsLoop) {
    LPDIRECTSOUNDBUFFER sb = dsb[inst][ch];
    if(sb) {
      sb->Stop();
      sb->SetCurrentPosition(0);
      sb->SetFrequency(DWORD(pow(2.0, float(pitch) / 12.0) * PITCH_BASE));
      sb->Play(0, 0, IsLoop ? DSBPLAY_LOOPING : 0);
    }
  }
  
  void Stop(DWORD ch, DWORD inst) {
    LPDIRECTSOUNDBUFFER sb = dsb[inst][ch];
    if(sb) {
      sb->Stop();
      sb->SetCurrentPosition(0);
    }
  }
}; //DirectSampler


//-----------------------------------------------------------------------------
// EP
//-----------------------------------------------------------------------------
int main(int argc, char *argv[]) {
  LPCSTR filename = argv[1];
  if(argc != 2) filename = "samples";
  DirectSampler dsampler;
  
  //Read sample file
  FILE *fp = fopen(filename, "rb");
  if(fp) {
    std::vector<char> buf;
    fseek(fp, 0, SEEK_END);
    int filesize = ftell(fp);
    fseek(fp, 0, SEEK_SET);
    buf.resize(filesize * 2);
    fread(&buf[0], 1, filesize, fp);
    fclose(fp);
    if(dsampler.Init(GetConsoleWindow(), &buf[0], filesize)) return 1;
  } else {
    printf("Can't open file %s\n", filename);
    return 1;
  }
  
  //Play
  static int offset      = 0x40;
  static DWORD  inst     = 0;
  static DWORD  max_inst = dsampler.NumSamples;
  printf("KEY -> [A,S,D,F]Oneshot, [Q,W,E,R]stop\n[DOWN]PitchDown, [UP]PitchUp, [LEFT]decrInst, [RIGHT]IncInst\n");
  while( !(GetAsyncKeyState(VK_ESCAPE) & 0x8000) ) {
    if( GetAsyncKeyState(VK_LEFT)  & 0x0001) inst--;
    if( GetAsyncKeyState(VK_RIGHT) & 0x0001) inst++;
    if( GetAsyncKeyState(VK_UP)    & 0x0001) offset++;
    if( GetAsyncKeyState(VK_DOWN)  & 0x0001) offset--;
    int iidx = inst % max_inst;
    if( GetAsyncKeyState('Q')      & 0x0001) dsampler.Stop(0, iidx);
    if( GetAsyncKeyState('W')      & 0x0001) dsampler.Stop(1, iidx);
    if( GetAsyncKeyState('E')      & 0x0001) dsampler.Stop(2, iidx);
    if( GetAsyncKeyState('R')      & 0x0001) dsampler.Stop(3, iidx);
    if( GetAsyncKeyState('A')      & 0x0001) dsampler.Play(0, iidx, offset + 0,  FALSE);
    if( GetAsyncKeyState('S')      & 0x0001) dsampler.Play(1, iidx, offset + 4,  FALSE);
    if( GetAsyncKeyState('D')      & 0x0001) dsampler.Play(2, iidx, offset + 7,  FALSE);
    if( GetAsyncKeyState('F')      & 0x0001) dsampler.Play(3, iidx, offset + 11, FALSE);
    printf("Inst:%d, Pitch:%d   \r", inst % max_inst, offset);
    Sleep(16);
  }
  dsampler.Term();
}

